from typing import Union, Optional
from pydantic import BaseModel, model_validator, root_validator
from typing_extensions import Literal

class ProsodyConfig(BaseModel):
    """
    Prosody configuration for Azure TTS.

    Args:
        rate: Speaking rate. Can be one of "x-slow", "slow", "medium", "fast", "x-fast", or a float. A float value of 1.0 represents normal speed.
        volume: Speaking volume. Can be one of "silent", "x-soft", "soft", "medium", "loud", "x-loud", or a float. A float value of 100 (x-loud) represents the highest volume and it's the default pitch.
        pitch: Speaking pitch. Can be one of "x-low", "low", "medium", "high", "x-high". The default pitch is "medium".
    """

    rate: Literal["x-slow", "slow", "medium", "fast", "x-fast"] | float | None = None
    volume: (
        Literal["silent", "x-soft", "soft", "medium", "loud", "x-loud"] | float | None
    ) = None
    pitch: Literal["x-low", "low", "medium", "high", "x-high"] | None = None


    @model_validator(mode="after")
    def validate_prosody(cls, instance: "ProsodyConfig") -> "ProsodyConfig":
        if instance.rate:
            if isinstance(instance.rate, float) and not 0.5 <= instance.rate <= 2:
                raise ValueError("Prosody rate must be between 0.5 and 2")
            if isinstance(instance.rate, str) and instance.rate not in [
                "x-slow",
                "slow",
                "medium",
                "fast",
                "x-fast",
            ]:
                raise ValueError(
                    "Prosody rate must be one of 'x-slow', 'slow', 'medium', 'fast', 'x-fast'"
                )
        if instance.volume:
            if isinstance(instance.volume, float) and not 0 <= instance.volume <= 100:
                raise ValueError("Prosody volume must be between 0 and 100")
            if isinstance(instance.volume, str) and instance.volume not in [
                "silent",
                "x-soft",
                "soft",
                "medium",
                "loud",
                "x-loud",
            ]:
                raise ValueError(
                    "Prosody volume must be one of 'silent', 'x-soft', 'soft', 'medium', 'loud', 'x-loud'"
                )

        if instance.pitch and instance.pitch not in [
            "x-low",
            "low",
            "medium",
            "high",
            "x-high",
        ]:
            raise ValueError(
                "Prosody pitch must be one of 'x-low', 'low', 'medium', 'high', 'x-high'"
            )

        return instance



import os
import importlib.util
import sys
import inspect
from jay_ai.config import (
    Agent,
    _default_on_agent_started_speaking,
    _default_on_user_started_speaking,
    _default_on_user_stopped_speaking,
    _default_on_agent_started_speaking,
    _default_on_agent_stopped_speaking,
    _default_on_user_speech_committed,
    _default_on_agent_speech_committed,
    _default_on_agent_speech_interrupted,
    _default_on_function_calls_collected,
    _default_on_function_calls_finished,
)


def load_agent(agent_path: str) -> Agent:
    agent_dir = os.path.dirname(agent_path)
    if agent_dir not in sys.path:
        sys.path.insert(0, agent_dir)

    try:
        functions_module = load_functions_module(agent_path)
    except Exception as e:
        print(f"Error loading functions module: {e}")
        sys.exit(1)

    # Build the list of tools from any functions that have _is_llm_function = True
    tools = []
    for name, member in inspect.getmembers(functions_module):
        if (
            inspect.isfunction(member)
            and getattr(member, "_is_llm_function", False) is True
        ):
            tools.append(member)

    # Required fields
    config = getattr(functions_module, "config", None)
    if not config:
        raise ValueError("Missing required top-level 'config = AgentConfig(id=...)'.")

    start_session = getattr(functions_module, "start_session", None)
    if not start_session:
        raise ValueError("Missing required top-level 'start_session' function.")

    agent_response_handler = getattr(functions_module, "agent_response_handler", None)
    if not agent_response_handler:
        raise ValueError(
            "Missing required top-level 'agent_response_handler' function."
        )

    # Optional fields, with defaults if missing
    on_user_started_speaking = getattr(
        functions_module, "on_user_started_speaking", _default_on_user_started_speaking
    )
    on_user_stopped_speaking = getattr(
        functions_module, "on_user_stopped_speaking", _default_on_user_stopped_speaking
    )
    on_agent_started_speaking = getattr(
        functions_module,
        "on_agent_started_speaking",
        _default_on_agent_started_speaking,
    )
    on_agent_stopped_speaking = getattr(
        functions_module,
        "on_agent_stopped_speaking",
        _default_on_agent_stopped_speaking,
    )
    on_user_speech_committed = getattr(
        functions_module, "on_user_speech_committed", _default_on_user_speech_committed
    )
    on_agent_speech_committed = getattr(
        functions_module,
        "on_agent_speech_committed",
        _default_on_agent_speech_committed,
    )
    on_agent_speech_interrupted = getattr(
        functions_module,
        "on_agent_speech_interrupted",
        _default_on_agent_speech_interrupted,
    )
    on_function_calls_collected = getattr(
        functions_module,
        "on_function_calls_collected",
        _default_on_function_calls_collected,
    )
    on_function_calls_finished = getattr(
        functions_module,
        "on_function_calls_finished",
        _default_on_function_calls_finished,
    )

    # Construct and return the Agent
    return Agent(
        id=config.id,
        start_session=start_session,
        agent_response_handler=agent_response_handler,
        tools=tools,
        on_user_started_speaking=on_user_started_speaking,
        on_user_stopped_speaking=on_user_stopped_speaking,
        on_agent_started_speaking=on_agent_started_speaking,
        on_agent_stopped_speaking=on_agent_stopped_speaking,
        on_user_speech_committed=on_user_speech_committed,
        on_agent_speech_committed=on_agent_speech_committed,
        on_agent_speech_interrupted=on_agent_speech_interrupted,
        on_function_calls_collected=on_function_calls_collected,
        on_function_calls_finished=on_function_calls_finished,
    )


def load_functions_module(module_path):
    spec = importlib.util.spec_from_file_location("", module_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module



from jay_ai.cli_types import LLMProvider, STTProvider, TTSProvider


AGENT_CONFIG_TEMPLATE = """import os
from openai import AsyncOpenAI
from jay_ai.cli_types import (
    VAD,
    STT,
    TTS,
    StartSessionInput,
    SessionConfig,
    AgentResponseHandlerInput,
    AgentConfig,
)

async def start_session(input: StartSessionInput):
    return SessionConfig(
        messages=[],
        vad=VAD.Silero(),
        stt={stt_injection},
        tts={tts_injection},
    )


async def agent_response_handler(input: AgentResponseHandlerInput):
    client = {llm_injection}
    messages = input["messages"]
    completion = await client.chat.completions.create(
        model="gpt-4o",
        messages=messages,
        stream=True,
    )
    return completion


config = AgentConfig(id="{agent_id}")
"""


def get_agent_template(
    agent_id: str,
    stt_provider: STTProvider,
    llm_provider: LLMProvider,
    tts_provider: TTSProvider,
) -> str:
    env_var_keys = tts_provider.value.env_var_keys + stt_provider.value.env_var_keys
    import_json = any(env_var.is_json for env_var in env_var_keys)

    file = AGENT_CONFIG_TEMPLATE.format(
        agent_id=agent_id,
        stt_injection=stt_provider.value.injection,
        llm_injection=llm_provider.value.injection,
        tts_injection=tts_provider.value.injection,
    )

    if import_json:
        file = "import json\n" + file

    return file



from enum import Enum
from typing import Dict, List, Any, Literal, Tuple, Union
from typing_extensions import TypedDict
from pydantic import BaseModel
from typing import List
from jay_ai.plugins import azure, elevenlabs, openai, google, cartesia, deepgram
from openai.types.chat import ChatCompletionUserMessageParam


class ProviderEnvVarKey(BaseModel):
    env_var_key: str
    is_json: bool


class ProviderInfo(BaseModel):
    label: str
    display_label: str
    injection: str | None
    env_var_keys: list[ProviderEnvVarKey]


class TTSProvider(Enum):
    ELEVENLABS = ProviderInfo(
        display_label="ElevenLabs",
        label="elevenlabs",
        injection='TTS.ElevenLabs(api_key=os.environ["ELEVENLABS_API_KEY"])',
        env_var_keys=[
            ProviderEnvVarKey(env_var_key="ELEVENLABS_API_KEY", is_json=False)
        ],
    )
    GOOGLE = ProviderInfo(
        display_label="Google",
        label="google",
        injection='TTS.Google(credentials=json.loads(os.environ["GOOGLE_CREDENTIALS"]))',
        env_var_keys=[
            ProviderEnvVarKey(env_var_key="GOOGLE_CREDENTIALS", is_json=True)
        ],
    )
    OPENAI = ProviderInfo(
        display_label="OpenAI",
        label="openai",
        injection='TTS.OpenAI(api_key=os.environ["OPENAI_API_KEY"])',
        env_var_keys=[ProviderEnvVarKey(env_var_key="OPENAI_API_KEY", is_json=False)],
    )
    AZURE = ProviderInfo(
        display_label="Azure",
        label="azure",
        injection='TTS.Azure(api_key=os.environ["AZURE_API_KEY"], region=os.environ["AZURE_REGION"])',
        env_var_keys=[
            ProviderEnvVarKey(env_var_key="AZURE_API_KEY", is_json=False),
            ProviderEnvVarKey(env_var_key="AZURE_REGION", is_json=False),
        ],
    )
    CARTESIA = ProviderInfo(
        display_label="Cartesia",
        label="cartesia",
        injection='TTS.Cartesia(api_key=os.environ["CARTESIA_API_KEY"])',
        env_var_keys=[ProviderEnvVarKey(env_var_key="CARTESIA_API_KEY", is_json=False)],
    )
    DEEPGRAM = ProviderInfo(
        display_label="Deepgram",
        label="deepgram",
        injection='TTS.Deepgram(api_key=os.environ["DEEPGRAM_API_KEY"])',
        env_var_keys=[ProviderEnvVarKey(env_var_key="DEEPGRAM_API_KEY", is_json=False)],
    )


class STTProvider(Enum):
    DEEPGRAM = ProviderInfo(
        display_label="Deepgram",
        label="deepgram",
        injection='STT.Deepgram(api_key=os.environ["DEEPGRAM_API_KEY"])',
        env_var_keys=[ProviderEnvVarKey(env_var_key="DEEPGRAM_API_KEY", is_json=False)],
    )
    OPENAI = ProviderInfo(
        display_label="OpenAI",
        label="openai",
        injection='STT.OpenAI(api_key=os.environ["OPENAI_API_KEY"])',
        env_var_keys=[ProviderEnvVarKey(env_var_key="OPENAI_API_KEY", is_json=False)],
    )
    AZURE = ProviderInfo(
        display_label="Azure",
        label="azure",
        injection='STT.Azure(api_key=os.environ["AZURE_API_KEY"], region=os.environ["AZURE_REGION"])',
        env_var_keys=[
            ProviderEnvVarKey(env_var_key="AZURE_API_KEY", is_json=False),
            ProviderEnvVarKey(env_var_key="AZURE_REGION", is_json=False),
        ],
    )


class VADProvider(Enum):
    SILERO = ProviderInfo(
        display_label="Silero", label="silero", injection=None, env_var_keys=[]
    )


class LLMProvider(Enum):
    OPENAI = ProviderInfo(
        display_label="OpenAI",
        label="openai",
        injection='AsyncOpenAI(api_key=os.environ["OPENAI_API_KEY"])',
        env_var_keys=[ProviderEnvVarKey(env_var_key="OPENAI_API_KEY", is_json=False)],
    )
    OPENAI_COMPATIBLE = ProviderInfo(
        display_label="OpenAI Compatible",
        label="openai_compatible",
        injection='AsyncOpenAI(api_key=os.environ["OPENAI_COMPATIBLE_API_KEY"])',
        env_var_keys=[
            ProviderEnvVarKey(env_var_key="OPENAI_COMPATIBLE_API_KEY", is_json=False)
        ],
    )


class Message(TypedDict):
    content: str
    role: Literal["system", "user", "assistant", "tool"]
    name: str | None = None
    tool_call_id: str | None = None


class TTS:
    class ElevenLabs(BaseModel):
        api_key: str
        voice: elevenlabs.Voice = elevenlabs.DEFAULT_VOICE
        model: elevenlabs.TTSModels | str = "eleven_turbo_v2_5"
        encoding: elevenlabs.TTSEncoding = "mp3_22050_32"
        streaming_latency: int = 3
        enable_ssml_parsing: bool = False
        chunk_length_schedule: list[int] = [80, 120, 200, 260]  # range is [50, 500]

    class OpenAI(BaseModel):
        api_key: str
        model: openai.TTSModels | str = "tts-1"
        voice: openai.TTSVoices | str = "alloy"
        speed: float = 1.0

    class Google(BaseModel):
        credentials: dict
        language: google.SpeechLanguages | str = "en-US"
        gender: google.Gender | str = "neutral"
        voice_name: str = ""  # Not required
        encoding: google.AudioEncoding | str = "linear16"
        sample_rate: int = 24000
        pitch: int = 0
        effects_profile_id: str = ""
        speaking_rate: float = 1.0

    class Azure(BaseModel):
        api_key: str
        region: str
        sample_rate: int = 24000
        voice: str | None = None
        language: str | None = None
        prosody: azure.ProsodyConfig | None = None
        endpoint_id: str | None = None

    class Deepgram(BaseModel):
        api_key: str
        model: str = "aura-asteria-en"
        encoding: str = "linear16"
        sample_rate: int = 24000

    class Cartesia(BaseModel):
        api_key: str
        model: cartesia.TTSModels | str = "sonic-english"
        language: cartesia.TTSLanguages | str = "en"
        encoding: cartesia.TTSEncoding = "pcm_s16le"
        voice: str | list[float] = cartesia.TTSDefaultVoiceId
        speed: cartesia.TTSVoiceSpeed | float | None = None
        emotion: list[cartesia.TTSVoiceEmotion | str] | None = None
        sample_rate: int = 24000


class STT:
    class Deepgram(BaseModel):
        api_key: str
        model: deepgram.DeepgramModels = "nova-2-general"
        language: deepgram.DeepgramLanguages = "en-US"
        interim_results: bool = True
        punctuate: bool = True
        smart_format: bool = True
        sample_rate: int = 16000
        no_delay: bool = True
        endpointing_ms: int = 25
        # enable filler words by default to improve turn detector accuracy
        filler_words: bool = True
        keywords: list[Tuple[str, float]] = []
        profanity_filter: bool = False

    class Azure(BaseModel):
        api_key: str
        region: str
        sample_rate: int = 16000
        num_channels: int = 1
        # Azure handles multiple languages and can auto-detect the language used. It requires the candidate set to be set.
        languages: list[str] = ["en-US"]

    class OpenAI(BaseModel):
        api_key: str
        language: str = "en"
        detect_language: bool = False
        model: openai.WhisperModels | str = "whisper-1"


class VAD:
    class Silero(BaseModel):
        min_speech_duration: float = 0.05
        min_silence_duration: float = 0.55
        prefix_padding_duration: float = 0.5
        max_buffered_speech: float = 60.0
        activation_threshold: float = 0.5
        sample_rate: Literal[8000, 16000] = 16000


class AgentConfig(BaseModel):
    id: str


class SessionConfig(BaseModel):
    messages: List[Message]
    vad: VAD.Silero
    stt: Union[STT.OpenAI, STT.Azure, STT.Deepgram]
    tts: Union[
        TTS.OpenAI,
        TTS.ElevenLabs,
        TTS.Google,
        TTS.Azure,
        TTS.Deepgram,
        TTS.Cartesia,
    ]
    custom_data: dict[str, Any] = {}
    first_message: str | None = None
    allow_interruptions: bool = True
    interrupt_speech_duration: float = 0.5
    interrupt_min_words: int = 0
    min_endpointing_delay: float = 0.5
    max_nested_fnc_calls: int = 1


class Function(TypedDict):
    arguments: str
    """
    The arguments to call the function with, as generated by the model in JSON
    format. Note that the model does not always generate valid JSON, and may
    hallucinate parameters not defined by your function schema. Validate the
    arguments in your code before calling your function.
    """

    name: str
    """The name of the function to call."""


class FunctionCallInput(TypedDict):
    id: str
    """The ID of the tool call."""

    function: Function
    """The function that the model called."""

    type: Literal["function"]
    """The type of the tool. Currently, only `function` is supported."""


class FunctionCallResult(TypedDict):
    content: str
    """The contents of the tool message."""

    role: Literal["tool"]
    """The role of the messages author, in this case `tool`."""

    function: Function
    """The function that the model called."""

    tool_call_id: str
    """Tool call that this message is responding to."""


class StartSessionInput(TypedDict):
    custom_data: dict


class AgentResponseHandlerInput(TypedDict):
    messages: List[Message]
    custom_data: dict[str, Any]


class OnUserStartedSpeakingInput(TypedDict):
    custom_data: dict[str, Any]


class OnUserStoppedSpeakingInput(TypedDict):
    custom_data: dict[str, Any]


class OnAgentStartedSpeakingInput(TypedDict):
    custom_data: dict[str, Any]


class OnAgentStoppedSpeakingInput(TypedDict):
    custom_data: dict[str, Any]


class OnUserSpeechCommittedInput(TypedDict):
    message: Message
    custom_data: dict[str, Any]


class OnAgentSpeechCommittedInput(TypedDict):
    message: Message
    custom_data: dict[str, Any]


class OnAgentSpeechInterruptedInput(TypedDict):
    message: Message
    custom_data: dict[str, Any]


class OnFunctionCallsCollectedInput(TypedDict):
    function_calls: List[FunctionCallInput]
    custom_data: dict[str, Any]


class OnFunctionCallsFinishedInput(TypedDict):
    results: List[FunctionCallResult]
    custom_data: dict[str, Any]


class AgentResponseHandlerPayload(BaseModel):
    custom_data: dict[str, Any]
    messages: List[Dict[str, Any]]


class StartSessionPayload(BaseModel):
    custom_data: dict[str, Any]
    agent_api_url: str


class OnUserStartedSpeakingPayload(BaseModel):
    custom_data: dict[str, Any]


class OnUserStoppedSpeakingPayload(BaseModel):
    custom_data: dict[str, Any]


class OnAgentStartedSpeakingPayload(BaseModel):
    custom_data: dict[str, Any]


class OnAgentStoppedSpeakingPayload(BaseModel):
    custom_data: dict[str, Any]


class OnUserSpeechCommittedPayload(BaseModel):
    custom_data: dict[str, Any]
    message: Message


class OnAgentSpeechCommittedPayload(BaseModel):
    custom_data: dict[str, Any]
    message: Message


class OnAgentSpeechInterruptedPayload(BaseModel):
    custom_data: dict[str, Any]
    message: Message


class OnFunctionCallsCollectedPayload(BaseModel):
    custom_data: dict[str, Any]
    function_calls: List[FunctionCallInput]


class OnFunctionCallsFinishedPayload(BaseModel):
    custom_data: dict[str, Any]
    results: List[FunctionCallResult]


class ToolCallPayload(BaseModel):
    function_arguments: dict[str, Any]
